import os
import unittest

from dotenv import load_dotenv
import requests

from src.chargily_pay.api import ChargilyClient
from src.chargily_pay.settings import CHARGILIY_TEST_URL
from src.chargily_pay.entity import (
    CheckoutItem,
    Customer,
    Address,
    Product,
    Price,
    Checkout,
    PaymentLink,
    PaymentItem,
)


class TestChargilyClient(unittest.TestCase):
    def setUp(self):
        load_dotenv()
        key = os.getenv("CHARGILY_KEY")
        secret = os.getenv("CHARGILY_SECRET")
        self.chargily = ChargilyClient(key, secret, url=CHARGILIY_TEST_URL)

    def test_get_balence(self):
        response = self.chargily.get_balance()
        self.assertEqual(type(response), dict)

    def test_create_customer(self):
        customer = Customer(
            name="Username",
            email="example@gmail.com",
            address=Address(address="Address", state="State", country="dz"),
        )
        response = self.chargily.create_customer(customer)
        self.assertEqual(type(response), dict)

    def test_create_customer_with_exception(self):
        customer = Customer(
            name="Username",
            email="example@gmail.com",
            address=Address(
                address="Address", state="State", country="wrong country code"
            ),
        )
        try:
            response = self.chargily.create_customer(customer)
        except requests.exceptions.HTTPError as e:
            self.assertEqual(type(e.response.json()), dict)

    def test_create_customer_with_phone(self):
        customer = Customer(
            name="Username",
            email="example@gmail.com",
            phone="12345678",
            address=Address(address="Address", state="State", country="dz"),
        )
        response = self.chargily.create_customer(customer)
        self.assertEqual(type(response), dict)
        self.assertEqual(response["phone"], "12345678")

    def test_create_customer_without_address(self):
        customer = Customer(
            name="Username",
            email="example@gmail.com",
        )
        response = self.chargily.create_customer(customer)
        self.assertEqual(type(response), dict)

    def test_update_customer(self):
        customer = Customer(
            name="Username",
            email="example@gmail.com",
        )
        response = self.chargily.create_customer(customer)
        customer_id = response["id"]
        self.assertEqual(response["name"], "Username")
        customer.name = "Username2"
        response = self.chargily.update_customer(customer_id, customer)
        self.assertEqual(response["name"], "Username2")

    def test_get_customer(self):
        customer = Customer(name="Username", email="example@gmail.com")
        response = self.chargily.create_customer(customer)
        customer_id = response["id"]
        response = self.chargily.retrieve_customer(customer_id)
        self.assertEqual(response["name"], "Username")

    def test_delete_customer(self):
        customer = Customer(name="Username", email="example@gmail.com")
        response = self.chargily.create_customer(customer)
        customer_id = response["id"]
        response = self.chargily.delete_customer(customer_id)
        try:
            response = self.chargily.retrieve_customer(customer_id)
        except requests.exceptions.HTTPError as err:
            self.assertEqual(err.response.status_code, 404)

    def test_list_customer(self):
        response = self.chargily.list_customers()
        self.assertEqual(type(response), dict)
        self.assertEqual(type(response["data"]), list)

    # Products

    def test_create_product(self):
        product1 = Product(
            name="Product name",
        )
        response = self.chargily.create_product(product1)
        self.assertEqual(type(response), dict)

        # product4 = Product(
        #     name="Product name",
        #     description="Product description",
        #     images=["https://example.com/image.png"], # TODO: there is a bug when upload image
        #     metadata=[],
        # )

        # response = self.chargily.create_product(product4)
        # self.assertEqual(type(response), dict)

    def test_update_product(self):
        product = Product(
            name="Product name",
            description="Product description",
        )
        response = self.chargily.create_product(product)
        product_id = response["id"]
        self.assertEqual(response["name"], "Product name")
        product.name = "Product name 2"
        response = self.chargily.update_product(product_id, product)
        self.assertEqual(response["name"], "Product name 2")

    def test_get_product(self):
        product = Product(
            name="Product name",
            description="Product description",
        )
        response = self.chargily.create_product(product)
        product_id = response["id"]
        response = self.chargily.retrieve_product(product_id)
        self.assertEqual(response["name"], "Product name")

    def test_delete_product(self):
        product = Product(
            name="Product name",
            description="Product description",
        )
        response = self.chargily.create_product(product)
        product_id = response["id"]
        response = self.chargily.delete_product(product_id)
        try:
            response = self.chargily.retrieve_product(product_id)
        except requests.exceptions.HTTPError as err:
            self.assertEqual(err.response.status_code, 404)

    def test_list_products(self):
        response = self.chargily.list_products()
        self.assertEqual(type(response), dict)

        response = self.chargily.list_products(per_page=1, page=1)
        self.assertEqual(type(response), dict)

    def test_retrieve_product_prices(self):
        product = Product(
            name="Product name",
            description="Product description",
        )
        response = self.chargily.create_product(product)
        product_id = response["id"]
        response = self.chargily.retrieve_product_prices(product_id)
        self.assertEqual(type(response), dict)
        self.assertEqual(response["data"], [])

        self.chargily.create_price(
            Price(amount=100, currency="dzd", product_id=product_id)
        )
        response = self.chargily.retrieve_product_prices(product_id)
        self.assertEqual(type(response), dict)
        self.assertEqual(len(response["data"]), 1)
        self.assertEqual(response["data"][0]["amount"], 100)

    # Price

    def test_create_price(self):
        product = Product(
            name="Product name",
            description="Product description",
        )
        response = self.chargily.create_product(product)
        product_id = response["id"]
        price = self.chargily.create_price(
            Price(amount=100, currency="dzd", product_id=product_id)
        )
        self.assertEqual(type(price), dict)

    def test_update_price(self):
        product = Product(
            name="Product name",
            description="Product description",
        )
        response = self.chargily.create_product(product)
        product_id = response["id"]

        # create price
        price = Price(amount=100, currency="dzd", product_id=product_id)
        response = self.chargily.create_price(price)

        # update price
        price_id = response["id"]
        self.assertEqual(response["amount"], 100)

        price = self.chargily.update_price(price_id, [{"value": "value"}])
        self.assertIsNotNone(price["metadata"])
        self.assertEqual(price["metadata"][0]["value"], "value")

    def test_retrieve_price(self):
        product = Product(
            name="Product name",
            description="Product description",
        )
        response = self.chargily.create_product(product)
        product_id = response["id"]
        price = self.chargily.create_price(
            Price(amount=100, currency="dzd", product_id=product_id)
        )
        price_id = price["id"]
        price = self.chargily.retrieve_price(price_id)
        self.assertEqual(type(price), dict)

    def test_list_prices(self):
        response = self.chargily.list_prices()
        self.assertEqual(type(response), dict)

        response = self.chargily.list_prices(per_page=1, page=1)
        self.assertEqual(type(response), dict)

    # Create checkout
    def test_create_checkout(self):
        product = Product(
            name="Product name",
            description="Product description",
        )
        response = self.chargily.create_product(product)
        product_id = response["id"]
        price = self.chargily.create_price(
            Price(amount=400, currency="dzd", product_id=product_id)
        )
        price_id = price["id"]
        checkout = self.chargily.create_checkout(
            Checkout(
                items=[CheckoutItem(price=price_id, quantity=2)],
                success_url="https://example.com/success",
                failure_url="https://example.com/failure",
                payment_method="cib",
            )
        )

        self.assertEqual(type(checkout), dict)

    def test_create_checkout_with_webhook(self):
        product = Product(
            name="Product name",
            description="Product description",
        )
        response = self.chargily.create_product(product)
        product_id = response["id"]
        price = self.chargily.create_price(
            Price(amount=100, currency="dzd", product_id=product_id)
        )
        price_id = price["id"]
        checkout = self.chargily.create_checkout(
            Checkout(
                items=[{"price": price_id, "quantity": 1}],
                success_url="https://example.com/success",
                failure_url="https://example.com/failure",
                webhook_endpoint="https://example.com/webhook",
            )
        )
        self.assertEqual(type(checkout), dict)
        self.assertEqual(checkout["webhook_endpoint"], "https://example.com/webhook")

    def test_create_checkout_with_payment_method(self):
        product = Product(
            name="Product name",
            description="Product description",
        )
        response = self.chargily.create_product(product)
        product_id = response["id"]
        price = self.chargily.create_price(
            Price(amount=100, currency="dzd", product_id=product_id)
        )
        price_id = price["id"]
        checkout = self.chargily.create_checkout(
            Checkout(
                items=[{"price": price_id, "quantity": 1}],
                success_url="https://example.com/success",
                payment_method="cib",
            )
        )
        self.assertEqual(type(checkout), dict)
        self.assertEqual(checkout["payment_method"], "cib")

    def test_create_checkout_with_amount(self):
        checkout = self.chargily.create_checkout(
            Checkout(
                amount=1000,
                currency="dzd",
                success_url="https://example.com/success",
            )
        )
        self.assertEqual(type(checkout), dict)

    def test_create_checkout_with_customer(self):
        product = Product(
            name="Product name",
            description="Product description",
        )
        response = self.chargily.create_product(product)
        product_id = response["id"]
        price = self.chargily.create_price(
            Price(amount=100, currency="dzd", product_id=product_id)
        )
        price_id = price["id"]
        customer = Customer(
            name="Username",
            email="customer@gmail.com",
        )
        response = self.chargily.create_customer(customer)
        customer_id = response["id"]

        checkout = self.chargily.create_checkout(
            Checkout(
                items=[{"price": price_id, "quantity": 1}],
                success_url="https://example.com/success",
                failure_url="https://example.com/failure",
                customer_id=customer_id,
            )
        )
        self.assertEqual(type(checkout), dict)

    def test_retrieve_checkout(self):
        product = Product(
            name="Product name",
            description="Product description",
        )
        response = self.chargily.create_product(product)
        product_id = response["id"]
        price = self.chargily.create_price(
            Price(amount=100, currency="dzd", product_id=product_id)
        )
        price_id = price["id"]
        checkout = self.chargily.create_checkout(
            Checkout(
                items=[{"price": price_id, "quantity": 1}],
                success_url="https://example.com/success",
                failure_url="https://example.com/failure",
            )
        )
        checkout_id = checkout["id"]
        checkout = self.chargily.retrieve_checkout(checkout_id)
        self.assertEqual(type(checkout), dict)

    def test_list_checkouts(self):
        response = self.chargily.list_checkouts()
        self.assertEqual(type(response), dict)

        response = self.chargily.list_checkouts(per_page=1, page=1)
        self.assertEqual(type(response), dict)

    def test_retrieve_checkout_items(self):
        product = Product(
            name="Product name",
            description="Product description",
        )
        response = self.chargily.create_product(product)
        product_id = response["id"]
        price = self.chargily.create_price(
            Price(amount=100, currency="dzd", product_id=product_id)
        )
        price_id = price["id"]
        checkout = self.chargily.create_checkout(
            Checkout(
                items=[{"price": price_id, "quantity": 1}],
                success_url="https://example.com/success",
                failure_url="https://example.com/failure",
            )
        )
        checkout_id = checkout["id"]
        checkout = self.chargily.retrieve_checkout_items(checkout_id)
        self.assertEqual(type(checkout), dict)

    def test_expire_checkout(self):
        product = Product(
            name="Product name",
            description="Product description",
        )
        response = self.chargily.create_product(product)
        product_id = response["id"]
        price = self.chargily.create_price(
            Price(amount=100, currency="dzd", product_id=product_id)
        )
        price_id = price["id"]
        checkout = self.chargily.create_checkout(
            Checkout(
                items=[{"price": price_id, "quantity": 1}],
                success_url="https://example.com/success",
                failure_url="https://example.com/failure",
            )
        )
        checkout_id = checkout["id"]
        checkout = self.chargily.expire_checkout(checkout_id)
        self.assertEqual(type(checkout), dict)
        self.assertEqual(checkout["status"], "expired")

    # Payment links

    def test_create_payment_link(self):
        product = Product(
            name="Product name",
            description="Product description",
        )
        response = self.chargily.create_product(product)
        product_id = response["id"]

        price = Price(amount=100, currency="dzd", product_id=product_id)
        response = self.chargily.create_price(price)
        price_id = response["id"]

        payment_link = self.chargily.create_payment_link(
            PaymentLink(
                name="Payment link name",
                items=[
                    PaymentItem(
                        price=price_id,
                        quantity=1,
                    )
                ],
            )
        )
        self.assertEqual(type(payment_link), dict)

    def test_update_payment_link(self):
        product = Product(
            name="Product name",
            description="Product description",
        )
        response = self.chargily.create_product(product)
        product_id = response["id"]

        price = Price(amount=100, currency="dzd", product_id=product_id)
        response = self.chargily.create_price(price)
        price_id = response["id"]
        payment_link = PaymentLink(
            name="Payment link name",
            items=[
                PaymentItem(
                    price=price_id,
                    quantity=1,
                )
            ],
        )
        response = self.chargily.create_payment_link(payment_link)
        payment_link_id = response["id"]
        self.assertEqual(response["name"], "Payment link name")
        try:
            payment_link.name = "Payment link name 2"
            response = self.chargily.update_payment_link(payment_link_id, payment_link)
            self.assertEqual(response["name"], "Payment link name 2")
            raise Exception("Should fail")
        except requests.exceptions.HTTPError as err:
            self.assertEqual(err.response.status_code, 422)

    def test_retrieve_payment_link(self):
        product = Product(
            name="Product name",
            description="Product description",
        )
        response = self.chargily.create_product(product)
        product_id = response["id"]

        price = Price(amount=100, currency="dzd", product_id=product_id)
        response = self.chargily.create_price(price)
        price_id = response["id"]
        payment_link = PaymentLink(
            name="Payment link name",
            items=[
                PaymentItem(
                    price=price_id,
                    quantity=1,
                )
            ],
        )
        response = self.chargily.create_payment_link(payment_link)
        payment_link_id = response["id"]
        response = self.chargily.retrieve_payment_link(payment_link_id)
        self.assertEqual(type(response), dict)

    def test_list_payment_links(self):
        response = self.chargily.list_payment_links()
        self.assertEqual(type(response), dict)

        response = self.chargily.list_payment_links(per_page=1, page=1)
        self.assertEqual(type(response), dict)



import hmac
import hashlib
from dataclasses import asdict

import requests
from requests.compat import urljoin

from .entity import Checkout, Customer, PaymentLink, Price, Product
from .settings import CHARGILIY_URL


# drop None values
exclude_none_value = lambda x: {k: v for (k, v) in x if v is not None}

asdict_true_value = lambda x: asdict(x, dict_factory=exclude_none_value)


def response_or_exception(fn):
    from functools import wraps

    @wraps(fn)
    def wrapper(*args, **kwargs):
        response: requests.Response = fn(*args, **kwargs)
        if response.status_code == 422:
            raise requests.exceptions.HTTPError(response, response=response)
        response.raise_for_status()

        return response.json()

    return wrapper


class ChargilyClient:
    def __init__(self, key, secret, url=CHARGILIY_URL):
        self.key = key
        self.url = url
        self.secret = secret
        self.headers = {
            "Authorization": f"Bearer {self.secret}",
            "Content-Type": "application/json",
        }

    # ==================================
    # Balance
    # ==================================

    def get_balance(self):
        """Get your balance"""
        response = requests.get(urljoin(self.url, "balance"), headers=self.headers)

        return response.json()

    # ==================================
    # Customers
    # ==================================
    @response_or_exception
    def create_customer(self, customer: Customer, *args, **kwargs):
        """Create a customer"""
        customer_dict = asdict_true_value(customer)
        response = requests.post(
            urljoin(self.url, "customers"), headers=self.headers, json=customer_dict
        )
        return response

    @response_or_exception
    def update_customer(self, id, customer: Customer):
        """Update a customer"""
        customer_dict = asdict_true_value(customer)
        response = requests.post(
            urljoin(self.url, f"customers/{id}"),
            headers=self.headers,
            json=customer_dict,
        )

        return response

    @response_or_exception
    def retrieve_customer(self, id):
        """Retrieve a customer"""
        response = requests.get(
            urljoin(self.url, f"customers/{id}"), headers=self.headers
        )
        return response

    @response_or_exception
    def list_customers(self, per_page: int = 10, page: int = 1):
        """List customers"""
        response = requests.get(
            urljoin(self.url, f"customers?page={page}"),
            headers=self.headers,
            params={"per_page": per_page},
        )

        return response

    @response_or_exception
    def delete_customer(self, id):
        """Delete a customer"""
        response = requests.delete(
            urljoin(self.url, f"customers/{id}"), headers=self.headers
        )

        return response

    # ==================================
    # Products
    # ==================================
    @response_or_exception
    def create_product(self, product: Product):
        """Create a product"""
        product_dict = asdict_true_value(product)

        response = requests.post(
            urljoin(self.url, "products"),
            headers=self.headers,
            json=product_dict,
        )

        return response

    @response_or_exception
    def update_product(self, id, product: Product):
        """Update a product"""
        product_dict = asdict_true_value(product)

        response = requests.post(
            urljoin(self.url, f"products/{id}"),
            headers=self.headers,
            json=product_dict,
        )

        return response

    @response_or_exception
    def retrieve_product(self, id):
        """Retrieve a product"""
        response = requests.get(
            urljoin(self.url, f"products/{id}"), headers=self.headers
        )

        return response

    @response_or_exception
    def list_products(self, per_page: int = 10, page: int = 1):
        """List products"""
        response = requests.get(
            urljoin(self.url, f"products?page={page}"),
            headers=self.headers,
            params={"per_page": per_page},
        )

        return response

    def delete_product(self, id):
        """Delete a product"""
        response = requests.delete(
            urljoin(self.url, f"products/{id}"), headers=self.headers
        )

        return response

    # todo: retrieve product prices
    @response_or_exception
    def retrieve_product_prices(self, id, per_page: int = 10, page: int = 1):
        """Retrieve product prices"""
        response = requests.get(
            urljoin(self.url, f"products/{id}/prices?page={page}"),
            headers=self.headers,
            params={"per_page": per_page},
        )

        return response

    # ==================================
    # Prices
    # ==================================

    @response_or_exception
    def create_price(self, price: Price):
        """Create a price"""
        price_dict = asdict_true_value(price)
        response = requests.post(
            urljoin(self.url, "prices"),
            headers=self.headers,
            json=price_dict,
        )

        return response

    @response_or_exception
    def update_price(self, id, metadata: list[dict]):
        """Update a price"""

        response = requests.post(
            urljoin(self.url, f"prices/{id}"),
            headers=self.headers,
            json={"metadata": metadata},
        )

        return response

    @response_or_exception
    def retrieve_price(self, id):
        """Retrieve a price"""
        response = requests.get(urljoin(self.url, f"prices/{id}"), headers=self.headers)

        return response

    @response_or_exception
    def list_prices(self, per_page: int = 10, page: int = 1):
        """List prices"""
        response = requests.get(
            urljoin(self.url, f"prices?page={page}"),
            headers=self.headers,
            params={"per_page": per_page},
        )

        return response

    # ==================================
    # Checkouts
    # ==================================

    @response_or_exception
    def create_checkout(self, checkout: Checkout):
        """Create a checkout"""
        checkout_dict = asdict_true_value(checkout)
        response = requests.post(
            urljoin(self.url, "checkouts"),
            headers=self.headers,
            json=checkout_dict,
        )

        return response

    @response_or_exception
    def retrieve_checkout(self, id):
        """Retrieve a checkout"""
        response = requests.get(
            urljoin(self.url, f"checkouts/{id}"), headers=self.headers
        )

        return response

    @response_or_exception
    def list_checkouts(self, per_page: int = 10, page: int = 1):
        """List checkouts"""
        response = requests.get(
            urljoin(self.url, f"checkouts?page={page}"),
            headers=self.headers,
            params={"per_page": per_page},
        )

        return response

    @response_or_exception
    def retrieve_checkout_items(self, id, per_page: int = 10, page: int = 1):
        """List checkouts items"""
        response = requests.get(
            urljoin(self.url, f"checkouts/{id}/items?page={page}"),
            headers=self.headers,
            params={"per_page": per_page},
        )

        return response

    @response_or_exception
    def expire_checkout(self, id):
        """Expire a checkout"""
        response = requests.post(
            urljoin(self.url, f"checkouts/{id}/expire"), headers=self.headers
        )

        return response

    # ==================================
    # Payment Links
    # ==================================
    @response_or_exception
    def create_payment_link(self, payment_link: PaymentLink):
        """Create a payment link"""
        payment_link_dict = asdict_true_value(payment_link)
        response = requests.post(
            urljoin(self.url, "payment-links"),
            headers=self.headers,
            json=payment_link_dict,
        )

        return response

    @response_or_exception
    def update_payment_link(self, id, payment_link: PaymentLink):
        """Update a payment link"""
        payment_link_dict = asdict_true_value(payment_link)
        response = requests.post(
            urljoin(self.url, f"payment-links/{id}"),
            headers=self.headers,
            json=payment_link_dict,
        )

        return response

    @response_or_exception
    def retrieve_payment_link(self, id):
        """Retrieve a payment link"""
        response = requests.get(
            urljoin(self.url, f"payment-links/{id}"), headers=self.headers
        )

        return response

    @response_or_exception
    def list_payment_links(self, per_page: int = 10, page: int = 1):
        """List payment links"""
        response = requests.get(
            urljoin(self.url, f"payment-links?page={page}"),
            headers=self.headers,
            params={"per_page": per_page},
        )

        return response

    @response_or_exception
    def retrieve_payment_link_items(self, id, per_page: int = 10, page: int = 1):
        """List payment link items"""
        response = requests.get(
            urljoin(self.url, f"payment-links/{id}/items?page={page}"),
            headers=self.headers,
            params={"per_page": per_page},
        )

        return response

    # ==================================
    # Utils
    # ==================================

    def validate_signature(self, signature: str, payload: str):
        computed_signature = hmac.new(
            self.secret.encode("utf-8"),
            payload.encode("utf-8"),
            hashlib.sha256,
        ).hexdigest()

        if hmac.compare_digest(signature, computed_signature):
            return True
        return False



from dataclasses import dataclass, field
from typing import Optional


@dataclass
class Address:
    country: Optional[str] = ""
    state: Optional[str] = ""
    address: Optional[str] = ""


@dataclass
class Customer:
    name: str
    email: str
    phone: Optional[str] = None
    address: Address = None
    metadata: list = field(default_factory=list)


@dataclass
class Product:
    name: str
    description: Optional[str] = None
    images: list[str] = None
    metadata: list[dict] = field(default_factory=list)


@dataclass
class Price:
    amount: int
    currency: str
    product_id: str
    metadata: list[dict] = field(default_factory=list)


@dataclass
class CheckoutItem:
    price: str
    quantity: int


@dataclass
class Checkout:
    success_url: str
    items: Optional[CheckoutItem] = None
    amount: Optional[int] = None
    currency: str = None
    failure_url: str = None
    customer_id: str = None
    description: str = None
    locale: str = None
    payment_method: str = None
    webhook_endpoint: str = None
    pass_fees_to_customer: bool = None
    metadata: list[dict] = field(default_factory=list)

    def __post_init__(self):
        if not self.items and not self.amount:
            raise Exception("Either items or amount must be provided")

        if self.amount:
            if self.amount <= 10:
                raise Exception("amount should be great than 10 dzd")
            if not self.currency:
                raise Exception("Currency must be provided when amount is provided")


@dataclass
class PaymentItem:
    price: str
    quantity: int
    adjustable_quantity: bool = None


@dataclass
class PaymentLink:
    name: str
    items: list[PaymentItem]
    after_completion_message: str = None
    locale: str = None
    pass_fees_to_customer: bool = None
    metadata: list[dict] = None
